// src/y-socket-io/y-socket-io.js
import * as Y3 from "yjs";
import * as AwarenessProtocol from "y-protocols/awareness";
import * as promise2 from "lib0/promise";
import * as encoding2 from "lib0/encoding";
import * as decoding3 from "lib0/decoding";
import { assert } from "lib0/testing";

// src/y-socket-io/user.js
var User = class {
  /**
   * @param {string} room
   * @param {string} userid identifies the user globally.
   */
  constructor(room, userid) {
    this.room = room;
    this.initialRedisSubId = "0";
    this.subs = /* @__PURE__ */ new Set();
    this.userid = userid;
  }
};

// src/api.js
import * as Y2 from "yjs";
import * as redis from "redis";
import * as map from "lib0/map";
import * as decoding2 from "lib0/decoding";
import * as awarenessProtocol2 from "y-protocols/awareness";
import * as array2 from "lib0/array";
import * as random from "lib0/random";
import * as number from "lib0/number";
import * as promise from "lib0/promise";
import * as math from "lib0/math";

// src/protocol.js
import * as Y from "yjs";
import * as error from "lib0/error";
import * as encoding from "lib0/encoding";
import * as decoding from "lib0/decoding";
import * as array from "lib0/array";
import * as awarenessProtocol from "y-protocols/awareness";
import * as buffer from "lib0/buffer";
import * as logging from "lib0/logging";
var log = logging.createModuleLogger("@y/redis/protocol");
var messageSync = 0;
var messageAwareness = 1;
var messageSyncStep1 = 0;
var messageSyncStep2 = 1;
var messageSyncUpdate = 2;
var mergeMessages = (messages) => {
  if (messages.length < 2) {
    return messages;
  }
  const aw = new awarenessProtocol.Awareness(new Y.Doc());
  const updates = [];
  messages.forEach((m) => {
    const decoder = decoding.createDecoder(m);
    try {
      const messageType = decoding.readUint8(decoder);
      switch (messageType) {
        case messageSync: {
          const syncType = decoding.readUint8(decoder);
          if (syncType === messageSyncUpdate) {
            updates.push(decoding.readVarUint8Array(decoder));
          } else {
            error.unexpectedCase();
          }
          break;
        }
        case messageAwareness: {
          awarenessProtocol.applyAwarenessUpdate(aw, decoding.readVarUint8Array(decoder), null);
          break;
        }
        default: {
          error.unexpectedCase();
        }
      }
    } catch (e) {
      log(logging.RED, "Error parsing message", buffer.toBase64(m), e);
    }
  });
  const result = [];
  updates.length > 0 && result.push(encoding.encode((encoder) => {
    encoding.writeVarUint(encoder, messageSync);
    encoding.writeVarUint(encoder, messageSyncUpdate);
    encoding.writeVarUint8Array(encoder, Y.mergeUpdates(updates));
  }));
  aw.states.size > 0 && result.push(encoding.encode((encoder) => {
    encoding.writeVarUint(encoder, messageAwareness);
    encoding.writeVarUint8Array(
      encoder,
      awarenessProtocol.encodeAwarenessUpdate(
        aw,
        array.from(aw.getStates().keys())
      )
    );
  }));
  aw.destroy();
  return result;
};

// src/api.js
import * as env from "lib0/environment";
import * as logging2 from "lib0/logging";
var logWorker = logging2.createModuleLogger("@y/redis/api/worker");
var logApi = logging2.createModuleLogger("@y/redis/api");
var ydocUpdateCallback = env.getConf("ydoc-update-callback");
if (ydocUpdateCallback != null && ydocUpdateCallback.slice(-1) !== "/") {
  ydocUpdateCallback += "/";
}
var isSmallerRedisId = (a, b) => {
  const [a1, a2 = "0"] = a.split("-");
  const [b1, b2 = "0"] = b.split("-");
  const a1n = number.parseInt(a1);
  const b1n = number.parseInt(b1);
  return a1n < b1n || a1n === b1n && number.parseInt(a2) < number.parseInt(b2);
};
var extractMessagesFromStreamReply = (streamReply, prefix) => {
  const messages = /* @__PURE__ */ new Map();
  streamReply?.forEach((docStreamReply) => {
    const { room, docid } = decodeRedisRoomStreamName(docStreamReply.name.toString(), prefix);
    const docMessages = map.setIfUndefined(
      map.setIfUndefined(
        messages,
        room,
        map.create
      ),
      docid,
      () => ({ lastId: array2.last(docStreamReply.messages).id, messages: (
        /** @type {Array<Uint8Array>} */
        []
      ) })
    );
    docStreamReply.messages.forEach((m) => {
      if (m.message.m != null) {
        docMessages.messages.push(
          /** @type {Uint8Array} */
          m.message.m
        );
      }
    });
  });
  return messages;
};
var computeRedisRoomStreamName = (room, docid, prefix) => `${prefix}:room:${encodeURIComponent(room)}:${encodeURIComponent(docid)}`;
var decodeRedisRoomStreamName = (rediskey, expectedPrefix) => {
  const match = rediskey.match(/^(.*):room:(.*):(.*)$/);
  if (match == null || match[1] !== expectedPrefix) {
    throw new Error(`Malformed stream name! prefix="${match?.[1]}" expectedPrefix="${expectedPrefix}", rediskey="${rediskey}"`);
  }
  return { room: decodeURIComponent(match[2]), docid: decodeURIComponent(match[3]) };
};
var createApiClient = async (store, { redisPrefix, redisUrl }) => {
  const a = new Api(store, redisPrefix, redisUrl);
  await a.redis.connect();
  try {
    await a.redis.xGroupCreate(a.redisWorkerStreamName, a.redisWorkerGroupName, "0", { MKSTREAM: true });
  } catch (e) {
  }
  return a;
};
var Api = class {
  /**
   * @param {import('./storage.js').AbstractStorage} store
   * @param {string=} prefix
   * @param {string=} url
   */
  constructor(store, prefix = "y", url = env.ensureConf("ysr-redis")) {
    this.store = store;
    this.prefix = prefix;
    this.consumername = random.uuidv4();
    this.redisWorkerTimeout = number.parseInt(env.getConf("redis-task-timeout") || "1000");
    this.redisMinMessageLifetime = number.parseInt(env.getConf("redis-min-message-lifetime") || "60000");
    this.redisWorkerStreamName = this.prefix + ":worker";
    this.redisWorkerGroupName = this.prefix + ":worker";
    this._destroyed = false;
    this.redis = redis.createClient({
      url,
      // scripting: https://github.com/redis/node-redis/#lua-scripts
      scripts: {
        addMessage: redis.defineScript({
          NUMBER_OF_KEYS: 1,
          SCRIPT: `
            if redis.call("EXISTS", KEYS[1]) == 0 then
              redis.call("XADD", "${this.redisWorkerStreamName}", "*", "compact", KEYS[1])
            end
            redis.call("XADD", KEYS[1], "*", "m", ARGV[1])
          `,
          /**
           * @param {string} key
           * @param {Buffer} message
           */
          transformArguments(key, message) {
            return [key, message];
          },
          /**
           * @param {null} x
           */
          transformReply(x) {
            return x;
          }
        }),
        xDelIfEmpty: redis.defineScript({
          NUMBER_OF_KEYS: 1,
          SCRIPT: `
            if redis.call("XLEN", KEYS[1]) == 0 then
              redis.call("DEL", KEYS[1])
            end
          `,
          /**
           * @param {string} key
           */
          transformArguments(key) {
            return [key];
          },
          /**
           * @param {null} x
           */
          transformReply(x) {
            return x;
          }
        })
      }
    });
  }
  /**
   * @param {Array<{key:string,id:string}>} streams streamname-clock pairs
   * @return {Promise<Array<{ stream: string, messages: Array<Uint8Array>, lastId: string }>>}
   */
  async getMessages(streams) {
    if (streams.length === 0) {
      await promise.wait(50);
      return [];
    }
    const reads = await this.redis.xRead(
      redis.commandOptions({ returnBuffers: true }),
      streams,
      { BLOCK: 1e3, COUNT: 1e3 }
    );
    const res = [];
    reads?.forEach((stream) => {
      res.push({
        stream: stream.name.toString(),
        messages: mergeMessages(stream.messages.map((message) => message.message.m).filter((m) => m != null)),
        lastId: array2.last(stream.messages).id.toString()
      });
    });
    return res;
  }
  /**
   * @param {string} room
   * @param {string} docid
   * @param {Buffer} m
   */
  addMessage(room, docid, m) {
    if (m[0] === messageSync && m[1] === messageSyncStep2) {
      if (m.byteLength < 4) {
        return promise.resolve();
      }
      m[1] = messageSyncUpdate;
    }
    return this.redis.addMessage(computeRedisRoomStreamName(room, docid, this.prefix), m);
  }
  /**
   * @param {string} room
   * @param {string} docid
   */
  async getStateVector(room, docid = "/") {
    return this.store.retrieveStateVector(room, docid);
  }
  /**
   * @param {string} room
   * @param {string} docid
   */
  async getDoc(room, docid) {
    const ms = extractMessagesFromStreamReply(await this.redis.xRead(redis.commandOptions({ returnBuffers: true }), { key: computeRedisRoomStreamName(room, docid, this.prefix), id: "0" }), this.prefix);
    const docMessages = ms.get(room)?.get(docid) || null;
    if (docMessages?.messages) logApi(`processing messages of length: ${docMessages?.messages.length} in room: ${room}`);
    const docstate = await this.store.retrieveDoc(room, docid);
    const ydoc = new Y2.Doc();
    const awareness = new awarenessProtocol2.Awareness(ydoc);
    awareness.setLocalState(null);
    const now = performance.now();
    ydoc.transact(() => {
      if (docstate) {
        Y2.applyUpdateV2(ydoc, docstate.doc);
      }
      docMessages?.messages.forEach((m) => {
        const decoder = decoding2.createDecoder(m);
        switch (decoding2.readVarUint(decoder)) {
          case 0: {
            if (decoding2.readVarUint(decoder) === 2) {
              Y2.applyUpdate(ydoc, decoding2.readVarUint8Array(decoder));
            }
            break;
          }
          case 1: {
            awarenessProtocol2.applyAwarenessUpdate(awareness, decoding2.readVarUint8Array(decoder), null);
            break;
          }
        }
      });
    });
    logApi(`took ${performance.now() - now}ms to process messages for room: ${room}`);
    return { ydoc, awareness, redisLastId: docMessages?.lastId.toString() || "0", storeReferences: docstate?.references || null };
  }
  /**
   * @param {Object} opts
   * @param {number} [opts.blockTime]
   * @param {number} [opts.tryReclaimCount]
   * @param {number} [opts.tryClaimCount]
   */
  async consumeWorkerQueue({ blockTime = 1e3, tryReclaimCount = 5, tryClaimCount = 5 } = {}) {
    const tasks = [];
    if (tryReclaimCount > 0) {
      const reclaimedTasks = await this.redis.xAutoClaim(this.redisWorkerStreamName, this.redisWorkerGroupName, this.consumername, this.redisWorkerTimeout, "0", { COUNT: tryReclaimCount });
      reclaimedTasks.messages.forEach((m) => {
        const stream = m?.message.compact;
        stream && tasks.push({ stream, id: m?.id });
      });
    }
    if (tryClaimCount > 0) {
      const claimedTasks = await this.redis.xReadGroup(this.redisWorkerGroupName, this.consumername, { key: this.redisWorkerStreamName, id: ">" }, { COUNT: tryClaimCount - tasks.length, BLOCK: tasks.length === 0 ? blockTime : void 0 });
      claimedTasks?.forEach((task) => {
        task.messages.forEach((message) => {
          const stream = message.message.compact;
          stream && tasks.push({ stream, id: message.id });
        });
      });
    }
    tasks.length > 0 && logWorker("Accepted tasks ", { tasks });
    let reclaimCounts = 0;
    await promise.all(tasks.map(async (task) => {
      const streamlen = await this.redis.xLen(task.stream);
      if (streamlen === 0) {
        await this.redis.multi().xDelIfEmpty(task.stream).xDel(this.redisWorkerStreamName, task.id).exec();
        logWorker("Stream still empty, removing recurring task from queue ", { stream: task.stream });
      } else {
        reclaimCounts++;
        const { room, docid } = decodeRedisRoomStreamName(task.stream, this.prefix);
        const { ydoc, storeReferences, redisLastId } = await this.getDoc(room, docid);
        const lastId = math.max(number.parseInt(redisLastId.split("-")[0]), number.parseInt(task.id.split("-")[0]));
        await this.store.persistDoc(room, docid, ydoc);
        await promise.all([
          storeReferences ? this.store.deleteReferences(room, docid, storeReferences) : promise.resolve(),
          this.redis.multi().xTrim(task.stream, "MINID", lastId - this.redisMinMessageLifetime).xAdd(this.redisWorkerStreamName, "*", { compact: task.stream }).xDel(this.redisWorkerStreamName, task.id).exec()
        ]);
        logWorker("Compacted stream ", { stream: task.stream, taskId: task.id, newLastId: lastId - this.redisMinMessageLifetime });
        try {
          if (ydocUpdateCallback != null) {
            const formData = new FormData();
            formData.append("ydoc", new Blob([Y2.encodeStateAsUpdateV2(ydoc)]));
            const res = await fetch(new URL(room, ydocUpdateCallback), { body: formData, method: "PUT" });
            if (!res.ok) {
              console.error(`Issue sending data to YDOC_UPDATE_CALLBACK. status="${res.status}" statusText="${res.statusText}"`);
            }
          }
        } catch (e) {
          console.error(e);
        }
        ydoc.destroy();
      }
    }));
    return { tasks, reclaimCounts };
  }
  async destroy() {
    this._destroyed = true;
    try {
      await this.redis.quit();
    } catch (e) {
    }
  }
};
var createWorker = async (store, opts) => {
  const a = await createApiClient(store, opts);
  return new Worker(a);
};
var Worker = class {
  /**
   * @param {Api} client
   * @param {Object} opts
   * @param {number} [opts.blockTime]
   * @param {number} [opts.tryReclaimCount]
   * @param {number} [opts.tryClaimCount]
   */
  constructor(client, opts = {}) {
    this.client = client;
    logWorker("Created worker process ", { id: client.consumername, prefix: client.prefix, minMessageLifetime: client.redisMinMessageLifetime });
    (async () => {
      let prev = performance.now();
      while (!client._destroyed) {
        try {
          const { reclaimCounts } = await client.consumeWorkerQueue(opts);
          const now = performance.now();
          if (reclaimCounts === 0) {
            await promise.wait(client.redisWorkerTimeout);
          } else if (now - prev < client.redisWorkerTimeout) {
            await promise.wait(client.redisWorkerTimeout - (now - prev));
          }
          prev = now;
        } catch (e) {
          console.error(e);
        }
      }
      logWorker("Ended worker process ", { id: client.consumername });
    })();
  }
};

// src/subscriber.js
import * as map2 from "lib0/map";
import * as array3 from "lib0/array";
var run = async (subscriber) => {
  while (true) {
    try {
      const ms = await subscriber.client.getMessages(array3.from(subscriber.subs.entries()).map(([stream, s]) => ({ key: stream, id: s.id })));
      for (let i = 0; i < ms.length; i++) {
        const m = ms[i];
        const sub = subscriber.subs.get(m.stream);
        if (sub == null) continue;
        sub.id = m.lastId;
        if (sub.nextId != null) {
          sub.id = sub.nextId;
          sub.nextId = null;
        }
        sub.fs.forEach((f) => f(m.stream, m.messages));
      }
    } catch (e) {
      console.error(e);
    }
  }
};
var createSubscriber = async (store, opts) => {
  const client = await createApiClient(store, opts);
  return new Subscriber(client);
};
var Subscriber = class {
  /**
   * @param {api.Api} client
   */
  constructor(client) {
    this.client = client;
    this.subs = /* @__PURE__ */ new Map();
    run(this);
  }
  /**
   * @param {string} stream
   * @param {string} id
   */
  ensureSubId(stream, id) {
    const sub = this.subs.get(stream);
    if (sub != null && isSmallerRedisId(id, sub.id)) {
      sub.nextId = id;
    }
  }
  /**
   * @param {string} stream
   * @param {SubHandler} f
   */
  subscribe(stream, f) {
    const sub = map2.setIfUndefined(this.subs, stream, () => ({ fs: /* @__PURE__ */ new Set(), id: "0", nextId: null }));
    sub.fs.add(f);
    return {
      redisId: sub.id
    };
  }
  /**
   * @param {string} stream
   * @param {SubHandler} f
   */
  unsubscribe(stream, f) {
    const sub = this.subs.get(stream);
    if (sub) {
      sub.fs.delete(f);
      if (sub.fs.size === 0) {
        this.subs.delete(stream);
      }
    }
  }
  destroy() {
    this.subs.clear();
    this.client.destroy();
  }
};

// src/y-socket-io/y-socket-io.js
var YSocketIO = class {
  /**
   * @type {Server}
   * @private
   * @readonly
   */
  io;
  /**
   * @type {YSocketIOConfiguration}
   * @private
   * @readonly
   */
  configuration;
  /**
   * @type {Namespace | null}
   * @public
   */
  nsp = null;
  /**
   * @type {api.Api | null}
   * @private
   */
  client = null;
  /**
   * @type {Subscriber | null}
   * @private
   */
  subscriber = null;
  /**
   * @type {Map<string, string>}
   * @private
   * @readonly
   */
  namespaceStreamMap = /* @__PURE__ */ new Map();
  /**
   * @type {Map<string, string>}
   * @private
   * @readonly
   */
  streamNamespaceMap = /* @__PURE__ */ new Map();
  /**
   * @type {Map<string, Namespace>}
   * @private
   * @readonly
   */
  namespaceMap = /* @__PURE__ */ new Map();
  /**
   * YSocketIO constructor.
   * @constructor
   * @param {Server} io Server instance from Socket IO
   * @param {YSocketIOConfiguration} configuration The YSocketIO configuration
   */
  constructor(io, configuration) {
    this.io = io;
    this.configuration = configuration;
  }
  /**
   * YSocketIO initialization.
   *
   *  This method set ups a dynamic namespace manager for namespaces that match with the regular expression `/^\/yjs\|.*$/`
   *  and adds the connection authentication middleware to the dynamics namespaces.
   *
   *  It also starts socket connection listeners.
   * @param {import('../storage.js').AbstractStorage} store
   * @param {{ redisPrefix?: string, redisUrl?: string }=} opts
   * @public
   */
  async initialize(store, { redisUrl, redisPrefix = "y" } = {}) {
    const [client, subscriber] = await promise2.all([
      createApiClient(store, { redisUrl, redisPrefix }),
      createSubscriber(store, { redisUrl, redisPrefix })
    ]);
    this.client = client;
    this.subscriber = subscriber;
    this.nsp = this.io.of(/^\/yjs\|.*$/);
    this.nsp.use(async (socket, next) => {
      if (this.configuration.authenticate == null) return next();
      const user = await this.configuration.authenticate(socket);
      if (user) {
        socket.user = new User(this.getNamespaceString(socket.nsp), user.userid);
        return next();
      } else return next(new Error("Unauthorized"));
    });
    this.nsp.on("connection", async (socket) => {
      assert(this.client);
      assert(this.subscriber);
      if (!socket.user) throw new Error("user does not exist in socket");
      const namespace = this.getNamespaceString(socket.nsp);
      const stream = computeRedisRoomStreamName(
        namespace,
        "index",
        redisPrefix
      );
      if (!this.namespaceMap.has(namespace)) {
        this.namespaceMap.set(namespace, socket.nsp);
      }
      socket.user.subs.add(stream);
      socket.user.initialRedisSubId = subscriber.subscribe(
        stream,
        this.redisMessageSubscriber
      ).redisId;
      this.namespaceStreamMap.set(namespace, stream);
      this.streamNamespaceMap.set(stream, namespace);
      this.initSyncListeners(socket);
      this.initAwarenessListeners(socket);
      this.initSocketListeners(socket);
      const doc = await this.client.getDoc(namespace, "index");
      if (isSmallerRedisId(doc.redisLastId, socket.user.initialRedisSubId)) {
        this.subscriber.ensureSubId(stream, doc.redisLastId);
      }
      this.startSynchronization(socket, doc);
    });
    return { client, subscriber };
  }
  /**
   * This function initializes the socket event listeners to synchronize document changes.
   *
   *  The synchronization protocol is as follows:
   *  - A client emits the sync step one event (`sync-step-1`) which sends the document as a state vector
   *    and the sync step two callback as an acknowledgment according to the socket io acknowledgments.
   *  - When the server receives the `sync-step-1` event, it executes the `syncStep2` acknowledgment callback and sends
   *    the difference between the received state vector and the local document (this difference is called an update).
   *  - The second step of the sync is to apply the update sent in the `syncStep2` callback parameters from the server
   *    to the document on the client side.
   *  - There is another event (`sync-update`) that is emitted from the client, which sends an update for the document,
   *    and when the server receives this event, it applies the received update to the local document.
   *  - When an update is applied to a document, it will fire the document's "update" event, which
   *    sends the update to clients connected to the document's namespace.
   * @private
   * @param {Socket} socket The socket connection
   * @readonly
   */
  initSyncListeners = (socket) => {
    socket.on(
      "sync-step-1",
      async (stateVector, syncStep2) => {
        assert(this.client);
        const doc = await this.client.getDoc(
          this.getNamespaceString(socket.nsp),
          "index"
        );
        syncStep2(Y3.encodeStateAsUpdate(doc.ydoc, stateVector));
      }
    );
    socket.on("sync-update", (update) => {
      assert(this.client);
      const message = Buffer.from(update.slice(0, update.byteLength));
      this.client.addMessage(
        this.getNamespaceString(socket.nsp),
        "index",
        Buffer.from(this.toRedis("sync-update", message))
      );
    });
  };
  /**
   * This function initializes socket event listeners to synchronize awareness changes.
   *
   *  The awareness protocol is as follows:
   *  - A client emits the `awareness-update` event by sending the awareness update.
   *  - The server receives that event and applies the received update to the local awareness.
   *  - When an update is applied to awareness, the awareness "update" event will fire, which
   *    sends the update to clients connected to the document namespace.
   * @private
   * @param {Socket} socket The socket connection
   * @readonly
   */
  initAwarenessListeners = (socket) => {
    socket.on("awareness-update", (update) => {
      assert(this.client);
      const message = Buffer.from(update.slice(0, update.byteLength));
      this.client.addMessage(
        this.getNamespaceString(socket.nsp),
        "index",
        Buffer.from(this.toRedis("awareness-update", new Uint8Array(message)))
      );
    });
  };
  /**
   *  This function initializes socket event listeners for general purposes.
   * @private
   * @param {Socket} socket The socket connection
   * @readonly
   */
  initSocketListeners = (socket) => {
    socket.on("disconnect", async () => {
      assert(this.subscriber);
      if (!socket.user) return;
      for (const ns of socket.user.subs) {
        const stream = this.namespaceStreamMap.get(ns);
        const nsp = this.namespaceMap.get(ns);
        if (nsp?.sockets.size === 0 && stream) {
          this.subscriber.unsubscribe(stream, this.redisMessageSubscriber);
          this.namespaceStreamMap.delete(ns);
          this.streamNamespaceMap.delete(stream);
          this.namespaceMap.delete(ns);
        }
      }
    });
  };
  /**
   * This function is called when a client connects and it emit the `sync-step-1` and `awareness-update`
   * events to the client to start the sync.
   * @private
   * @param {Socket} socket The socket connection
   * @param {RedisDoc} doc The document
   */
  startSynchronization = (socket, doc) => {
    socket.emit(
      "sync-step-1",
      Y3.encodeStateVector(doc.ydoc),
      (update) => {
        assert(this.client);
        const message = Buffer.from(update.slice(0, update.byteLength));
        this.client.addMessage(
          this.getNamespaceString(socket.nsp),
          "index",
          Buffer.from(this.toRedis("sync-step-2", message))
        );
      }
    );
    if (doc.awareness.states.size > 0) {
      socket.emit(
        "awareness-update",
        AwarenessProtocol.encodeAwarenessUpdate(
          doc.awareness,
          Array.from(doc.awareness.getStates().keys())
        )
      );
    }
  };
  /**
   * @private
   * @param {string} stream
   * @param {Array<Uint8Array>} messages
   */
  redisMessageSubscriber = (stream, messages) => {
    const namespace = this.streamNamespaceMap.get(stream);
    if (!namespace) return;
    const nsp = this.namespaceMap.get(namespace);
    if (!nsp) return;
    if (nsp.sockets.size === 0 && this.subscriber) {
      this.subscriber.unsubscribe(stream, this.redisMessageSubscriber);
      this.namespaceStreamMap.delete(namespace);
      this.streamNamespaceMap.delete(stream);
      this.namespaceMap.delete(namespace);
    }
    const updates = [];
    const awareness = [];
    for (const m of messages) {
      const decoded = this.fromRedis(m);
      if (decoded.type === "awareness-update") awareness.push(decoded.message);
      else updates.push(decoded.message);
    }
    for (const msg of updates) {
      if (msg.length === 0) continue;
      nsp.emit("sync-update", msg);
    }
    for (const msg of awareness) {
      if (msg.length === 0) continue;
      nsp.emit("awareness-update", msg);
    }
  };
  /**
   * @param {Namespace} namespace
   */
  getNamespaceString(namespace) {
    return namespace.name.replace(/\/yjs\|/, "");
  }
  /**
   * @param {Uint8Array | ArrayBuffer | Buffer} message
   * @returns {{ type: EventType, message: Uint8Array }}
   */
  fromRedis(message) {
    const decoder = decoding3.createDecoder(new Uint8Array(message));
    const opt = decoding3.readUint8(decoder);
    switch (opt) {
      case messageSync: {
        const syncStep = decoding3.readUint8(decoder);
        const message2 = decoding3.readVarUint8Array(decoder);
        switch (syncStep) {
          case messageSyncStep1:
            return {
              type: "sync-step-1",
              message: message2
            };
          case messageSyncStep2:
            return {
              type: "sync-step-2",
              message: message2
            };
          case messageSyncUpdate:
            return {
              type: "sync-update",
              message: message2
            };
        }
        break;
      }
      case messageAwareness: {
        const message2 = decoding3.readVarUint8Array(decoder);
        return {
          type: "awareness-update",
          message: message2
        };
      }
    }
    throw new Error("unknown encoding");
  }
  /**
   * @param {EventType} type
   * @param {Uint8Array | ArrayBuffer | Buffer} message
   */
  toRedis(type, message) {
    return encoding2.encode((encoder) => {
      switch (type) {
        case "sync-step-1":
          encoding2.writeVarUint(encoder, messageSync);
          encoding2.writeVarUint(encoder, messageSyncStep1);
          break;
        case "sync-step-2":
          encoding2.writeVarUint(encoder, messageSync);
          encoding2.writeVarUint(encoder, messageSyncStep2);
          break;
        case "sync-update":
          encoding2.writeVarUint(encoder, messageSync);
          encoding2.writeVarUint(encoder, messageSyncUpdate);
          break;
        case "awareness-update":
          encoding2.writeVarUint(encoder, messageAwareness);
          break;
      }
      encoding2.writeVarUint8Array(encoder, new Uint8Array(message));
    });
  }
  destroy() {
    try {
      this.subscriber?.destroy();
      return this.client?.destroy();
    } catch (e) {
      console.error(e);
    }
  }
};

// src/socketio.js
var YSocketIOServer = class {
  /**
   * @param {YSocketIO} app
   * @param {import('./api.js').Api} client
   * @param {import('./subscriber.js').Subscriber} subscriber
   */
  constructor(app, client, subscriber) {
    this.app = app;
    this.subscriber = subscriber;
    this.client = client;
  }
  async destroy() {
    this.subscriber.destroy();
    await this.client.destroy();
  }
};
var registerYSocketIOServer = async (io, store, { authenticate, redisUrl, redisPrefix }) => {
  const app = new YSocketIO(io, { authenticate });
  const { client, subscriber } = await app.initialize(store, { redisUrl, redisPrefix });
  return new YSocketIOServer(app, client, subscriber);
};

// src/server.js
import * as env2 from "lib0/environment";
import * as logging3 from "lib0/logging";
import * as jwt from "lib0/crypto/jwt";
import * as ecdsa from "lib0/crypto/ecdsa";
import * as json from "lib0/json";
import { Server } from "socket.io";
import http from "http";
import { randomUUID } from "crypto";
var authPublicKey = env2.getConf("auth-public-key");
var wsServerPublicKey = null;
if (authPublicKey) {
  ecdsa.importKeyJwk(json.parse(env2.ensureConf("auth-public-key"))).then((key) => {
    wsServerPublicKey = key;
  });
}
var createYSocketIOServer = async ({
  redisPrefix = "y",
  port,
  store
}) => {
  const httpServer = http.createServer((req, res) => {
    res.writeHead(200, { "Content-Type": "application/json" });
    res.end(JSON.stringify({ ok: true }));
  });
  const io = new Server(httpServer);
  const server = await registerYSocketIOServer(io, store, {
    redisPrefix,
    authenticate: async (socket) => {
      if (!wsServerPublicKey) return { userid: randomUUID().toString() };
      const token = socket.handshake.query.yauth;
      if (!token) return null;
      const { payload: userToken } = await jwt.verifyJwt(
        wsServerPublicKey,
        typeof token === "string" ? token : token[0]
      );
      if (!userToken.yuserid) return null;
      return { userid: userToken.yuserid };
    }
  });
  httpServer.listen(port, void 0, void 0, () => {
    logging3.print(logging3.GREEN, "[y-redis] Listening to port ", port);
  });
  return server;
};

// src/storage.js
import * as Y4 from "yjs";
import * as err from "lib0/error";
var AbstractStorage = class {
  /**
   * @param {string} _room
   * @param {string} _docname
   * @param {Y.Doc} _ydoc
   * @return {Promise<void>}
   */
  persistDoc(_room, _docname, _ydoc) {
    err.methodUnimplemented();
  }
  /**
   * @param {string} _room
   * @param {string} _docname
   * @return {Promise<{ doc: Uint8Array, references: Array<any> }|null>}
   */
  retrieveDoc(_room, _docname) {
    err.methodUnimplemented();
  }
  /**
   * This can be implemented by the storage provider for better efficiency. The state vector must be
   * updated when persistDoc is called. Otherwise, we pull the ydoc and compute the state vector.
   *
   * @param {string} room
   * @param {string} docname
   * @return {Promise<Uint8Array|null>}
   */
  async retrieveStateVector(room, docname) {
    const r = await this.retrieveDoc(room, docname);
    return r ? Y4.encodeStateVectorFromUpdateV2(r.doc) : null;
  }
  /**
   * @param {string} _room
   * @param {string} _docname
   * @param {Array<any>} _storeReferences
   * @return {Promise<void>}
   */
  deleteReferences(_room, _docname, _storeReferences) {
    err.methodUnimplemented();
  }
  async destroy() {
  }
};
export {
  AbstractStorage,
  Api,
  Subscriber,
  User,
  Worker,
  YSocketIO,
  computeRedisRoomStreamName,
  createApiClient,
  createSubscriber,
  createWorker,
  createYSocketIOServer,
  isSmallerRedisId,
  registerYSocketIOServer
};
